# Import all decorators from 'decorators.py'
from decorators import *

# Necessary imports to access pandas functions and structures
from pandas import set_option, DataFrame, json_normalize, concat

# Mute chained assignment warnings
set_option('mode.chained_assignment', None)

# Necessary imports to access json functions
from json import loads

class JSON_DataFrame_Parser():
    """A class used to parse JSON files into normalized pandas DataFrames.
    """

    @entry_exit
    def __init__(self):
        """This function initializes the JSON_DataFrame_Parser object.
        """
        pass

    def flatten(self, data : DataFrame) -> DataFrame:
        """This function flattens a single user specified DataFrame by 1 level and returns the resulting DataFrame.

        Args:
            data (DataFrame): DataFrame formatted as [[key, value(s)]] to be flattened.

        Returns:
            DataFrame: Flattened 'data' DataFrame by 1 level.
        """

        # Initialize a copy of user specified data
        DATA = data.copy()

        # Convert user specified data to JSON format
        to_json_data = DATA.to_json(orient="records")
        json_loads_data = loads(to_json_data)

        # Flatten JSON data by 1 level
        flattened_data = json_normalize(json_loads_data, max_level=1)

        # Return flattened DataFrame
        return flattened_data

    def normalize(self, data : DataFrame, primary_key : DataFrame, foreign_key : DataFrame, columns_to_exclude : list[str] = []) -> list[DataFrame]:
        """This function normalizes a single user specified DataFrame by 1 level and returns the results as a list of DataFrames.

        Args:
            data (DataFrame): DataFrame formatted as [[key, value(s)]] to be normalized.
            primary_key (DataFrame): DataFrame formatted as [[primary_key_name]] containing a list of integers.
            foreign_key (DataFrame): DataFrame formatted as [[foreign_key_name]] containing a list of integers.
            columns_to_exclude (list[str], optional): List of columns to exclude from the normalization process. Defaults to [].

        Returns:
            list[DataFrame]: Normalized 'data' by 1 level.
        """    

        # Initialize a copy of user specified data
        DATA = data.copy()

        # Initialize an empty list of column names to be normalized
        columns_to_normalize = []

        # Initialize an empty list of column names to be exploded
        columns_to_explode = []

        # Iterate over all columns in the data
        for column in DATA.columns.drop(columns_to_exclude):
            
            # Iterate over all rows in the current column
            for row in DATA[column].index:

                # Determine whether the current column contains 'dict' values to be normalized
                if isinstance(DATA[column].loc[row], dict):
                    columns_to_normalize.append(DATA[[column]])

                    # Set foreign keys
                    DATA[column] = DATA.index

                    break

                # Determine whether the current column contains 'list' values to be exploded
                elif isinstance(DATA[column].loc[row], list):
                    columns_to_explode.append(DATA[[column]])

                    # Set foreign keys
                    DATA[column] = DATA.index

                    break

        # Set primary key for each column to be normalized
        for column in range(len(columns_to_normalize)):
            columns_to_normalize[column][primary_key.columns[0]] = primary_key

            # Sort columns
            columns_to_normalize[column] = columns_to_normalize[column][[primary_key.columns[0], columns_to_normalize[column].columns[0]]]

            # Fill 'nan' to empty dictionary
            columns_to_normalize[column][columns_to_normalize[column].columns[1]] = columns_to_normalize[column][columns_to_normalize[column].columns[1]].apply(lambda val: val if isinstance(val, dict) else {})

            # Flatten data by 1 level
            columns_to_normalize[column] = self.flatten(data=columns_to_normalize[column])

            # Set foreign keys
            columns_to_normalize[column]['foreign_key'] = DATA.index

            # Reorder columns
            columns_to_normalize[column] = columns_to_normalize[column][[f"{columns_to_normalize[column].columns[0]}", 'foreign_key', *list(columns_to_normalize[column].columns[1:])]]

        # Set primary key for each column to be exploded
        for column in range(len(columns_to_explode)):
            columns_to_explode[column][primary_key.columns[0]] = primary_key

            # Sort columns
            columns_to_explode[column] = columns_to_explode[column][[primary_key.columns[0], columns_to_explode[column].columns[0]]]

            # Fill 'nan' to list
            columns_to_explode[column][columns_to_explode[column].columns[1]] = columns_to_explode[column][columns_to_explode[column].columns[1]].apply(lambda val: val if isinstance(val, list) else [])

            # Explode data
            columns_to_explode[column] = self.explode(data=columns_to_explode[column], primary_key=primary_key, foreign_key=foreign_key)

        # Return list of normalized DataFrames
        return [DATA, columns_to_normalize, columns_to_explode]

    def explode(self, data : DataFrame, primary_key : DataFrame, foreign_key : DataFrame) -> DataFrame:
        """This function explodes a DataFrame, creating multiple columns to represent the exploded records.

        Args:
            data (DataFrame): DataFrame formatted as [[key, value(s)]] to be normalized.
            primary_key (DataFrame): DataFrame formatted as [[primary_key_name]] containing a list of integers.
            foreign_key (DataFrame): DataFrame formatted as [[foreign_key_name]] containing a list of integers.

        Returns:
            DataFrame: Exploded DataFrame.
        """        

        # Initialize a copy of user specified data
        DATA = data.copy()

        # Explode data into multiple columns
        exploded_data = DataFrame(DATA[DATA.columns[1]].to_list())

        # Rename columns in exploded data
        new_column_names = []
        for column in range(len(exploded_data.columns)):
            new_column_names.append(f"{DATA.columns[1]}.{column + 1}")
        exploded_data.columns = new_column_names

        # Attach primary and foreign keys
        exploded_data = concat([primary_key, foreign_key, exploded_data], axis=1)

        # Return exploded data
        return exploded_data